/* 
 * import.c
 * 
 * import C_SYNC data into the game database
 * 
 * Ken Stevens, 1996
 *
 * This program takes the output of the deity command "sync everything"
 * and imports the data in that format into the current game database.
 */

#if defined(aix) || defined(linux) || defined(solaris)
#include <unistd.h>
#endif /* aix or linux */

#include <sys/types.h>
#include <fcntl.h>
#include <sys/uio.h>
#include <sys/file.h>
#include <stdio.h>

#include <time.h>
#include "gamesdef.h"
#include "misc.h"
#include "player.h"
#include "edb.h"
#include "xy.h"
#include "nsc.h"
#include "version.h"
#include "file.h"
#include "sect.h"
#include "ship.h"
#include "plane.h"
#include "land.h"
#include "nuke.h"
#include "news.h"
#include "trade.h"
#include "commodity.h"
#include "nat.h"
#include "loan.h"
#include "map.h"

#ifndef SERVERTYPES
#define SERVERTYPES
#endif
#include "edbproc.h"

s_char	program[] = "file";

int verbose = 0;
int unconditional = 0;

/* translation table from old ship types to new types */
int *shipXlateCodes = 0;
int *planeXlateCodes = 0;
int *landXlateCodes = 0;
int *nukeXlateCodes = 0;

main(argc, argv)
	int argc;
	char *argv[];
{
	extern char *optarg;
	extern	s_char *datadir;
	int opt;
	char *config_file = NULL;
	char *data_file = NULL;
	s_char	buf[255];
	int	i;

	while ((opt = getopt (argc, argv, "e:f:huv")) != EOF) {
		switch (opt) {
			case 'e':
				config_file = optarg;
				break;
			case 'f':
				data_file = optarg;
				break;
			case 'h':
				usage();
				exit(0);
				break;
			case 'u':
				unconditional++;
				break;
			case 'v':
				verbose++;
				break;
		}
	}

	emp_config (config_file);

	if (access(datadir, F_OK) < 0 && mkdir(datadir, 0750) < 0) {
		perror(datadir);
		printf("Can't make game directory\n");
		exit(1);
	}
	if (!unconditional) {
		printf("WARNING: this blasts the existing game in %s (if any)\n", datadir);
		printf("continue? ");
		gets(buf);
		if (*buf != 'y' && *buf != 'Y')
			exit(1);
	}
	
	for (i=0; i < EF_MAX; i++) {
/*		if (ef_open(i, O_RDWR|O_CREAT|O_TRUNC, 0) < 0) { */
		if (ef_open(i, O_RDWR, 0) < 0) { 
			perror("ef_open");
			exit(1);
		}
	}

	process_csync(data_file);

	for (i=0; i < EF_MAX; i++) {
		ef_close(i);
	}

	exit(0);	
}

usage(void)
{
	printf("import [ -e configfile ] [ -f csync_data ] -h\n\n"
		"	-e configfile	to specify which empire config file\n"
		"	-f csync_data	to specify csync_data input file\n"
		"	-v	verbose mode\n"
		"	-u	unconditional (no prompts)\n\n");
	exit(0);
}

/*
 * Translation functions
 */

char sectorDes(Empire_Database *edb, int descode)
{
	/* given the old sector des code, lookup the mnemonic */

	struct dchrstr *bld = edb_getSectorBuild(edb, descode);

	return bld? bld->d_mnem & 0x7f : '?';
}

int newDes(char oldDes)
{
	/* given the mnemonic, lookup the sector designation code */
	/* it would be better to do this once and build a table...*/

	int i;
	for (i=0; i < SCT_MAXDEF; ++i)
		if (dchr[i].d_mnem == oldDes)
			return i;
	return 0;
}

char *shipTypeName(Empire_Database *edb, int code)
{
	struct mchrstr *bld = edb_getShipBuild(edb, code);
	return bld->m_name;
}

int newShipTypeName(char *name)
{
	extern int shp_maxno;
	int i;
	for (i=0; i < shp_maxno; ++i)
		if (strncmp(mchr[i].m_name,name,3)==0)
			return i;
	return -1;
}

char *landTypeName(Empire_Database *edb, int code)
{
	struct lchrstr *bld = edb_getLandBuild(edb, code);
	return bld->l_name;
}

int newLandTypeName(char *name)
{
	extern int lnd_maxno;
	int i;
	for (i=0; i < lnd_maxno; ++i)
		if (strncmp(lchr[i].l_name,name,3)==0)
			return i;
	return -1;
}

char *planeTypeName(Empire_Database *edb, int code)
{
	struct plchrstr *bld = edb_getPlaneBuild(edb, code);
	return bld->pl_name;
}

int newPlaneTypeName(char *name)
{
	extern int pln_maxno;
	int i;
	for (i=0; i < pln_maxno; ++i)
		if (strncmp(plchr[i].pl_name,name,3)==0)
			return i;
	return -1;
}

char *nukeTypeName(Empire_Database *edb, int code)
{
	struct nchrstr *bld = edb_getNukeBuild(edb, code);
	return bld->n_name;
}

int newNukeTypeName(char *name)
{
	extern int nuk_maxno;
	int i;
	for (i=0; i < nuk_maxno; ++i)
		if (strncmp(nchr[i].n_name,name,3)==0)
			return i;
	return -1;
}

void repairCoord(coord *x, coord *y)
{
	if (*x<0) *x += WORLD_X;
	if (*y<0) *y += WORLD_Y;
}

/*
 * Put functions.
 */

void ImportOneSector(void *context, void *object)
{
	Empire_Database *edb = (Empire_Database *) context;
	struct sctstr *sp= (struct sctstr *) object;

	sp->ef_type = EF_SECTOR;
	sp->sct_type = newDes(sectorDes(edb, sp->sct_type));
        repairCoord(&sp->sct_x, &sp->sct_y);
        repairCoord(&sp->sct_dist_x, &sp->sct_dist_y);
	putsect(sp);
}

void ImportOneUnit(void *context, void *object)
{
	Empire_Database *edb = (Empire_Database *)context;
	struct lndstr *lp = (struct lndstr *)object;
	int newtype = newLandTypeName(landTypeName(edb, lp->lnd_type));

	if (verbose && newtype<0) {
		printf("Cannot import unit #%d (type name = '%s' index = %d)\n",
			lp->lnd_uid, landTypeName(edb, lp->lnd_type), lp->lnd_type);
		return;
	}
	lp->lnd_type = newtype;
	lp->ef_type = EF_LAND;
	repairCoord(&lp->lnd_x, &lp->lnd_y);
	repairCoord(&lp->lnd_opx, &lp->lnd_opy);

	putland(lp->lnd_uid, lp);
}

void ImportOneShip(void *context, void *object)
{
	Empire_Database *edb = (Empire_Database *)context;
	struct shpstr *sp = (struct shpstr *)object;
	int i;
	int newtype = newShipTypeName(shipTypeName(edb, sp->shp_type));

	if (verbose && newtype<0) {
		printf("Cannot import ship #%d (%s)\n",
			sp->shp_uid, shipTypeName(edb, sp->shp_type) );
		return;
	}
	sp->shp_type = newtype;
	sp->ef_type = EF_SHIP;
	repairCoord(&sp->shp_x, &sp->shp_y);
	repairCoord(&sp->shp_opx, &sp->shp_opy);
	repairCoord(&sp->shp_orig_x, &sp->shp_orig_y);
	for (i=0;i<2;++i)
		repairCoord(&sp->shp_destx[i], &sp->shp_desty[i]);
	putship(sp->shp_uid, sp);
}

void ImportOnePlane(void *context, void *object)
{
	Empire_Database *edb = (Empire_Database *)context;
	struct plnstr *pp = (struct plnstr *)object;
	int newtype = newPlaneTypeName(planeTypeName(edb, pp->pln_type));

	if (verbose && newtype<0) {
		printf("Cannot import plane #%d (%s)\n",
			pp->pln_uid, planeTypeName(edb, pp->pln_type) );
		return;
	}
	pp->pln_type = newtype;
	pp->ef_type = EF_PLANE;
	repairCoord(&pp->pln_x, &pp->pln_y);
	repairCoord(&pp->pln_opx, &pp->pln_opy);
	putplane(pp->pln_uid, pp);
}

void ImportOneNuke(void *context, void *object)
{
	Empire_Database *edb = (Empire_Database *)context;
	struct nukstr *np = (struct nukstr *)object;

	/* Ken: need to do code to translate nuke types... */
	
	np->ef_type = EF_NUKE;
	repairCoord(&np->nuk_x, &np->nuk_y);
	putnuke(np->nuk_uid, np);
}

void ImportOneNation(void *context, void *object)
{
	Empire_Database *edb = (Empire_Database *)context;
	struct natstr *natp = (struct natstr *)object;

	natp->ef_type = EF_NATION;

	/* Ken: this is a hack until we get nat_stat passed in CSYNC data... */
	
	if (natp->nat_cnum) 
		natp->nat_stat = STAT_NORM|STAT_INUSE;
	else		
		natp->nat_stat = STAT_GOD|STAT_NORM|STAT_INUSE;

	/* this is necessary because the server uses a different range */
	repairCoord(&natp->nat_xstart, &natp->nat_ystart);
	repairCoord(&natp->nat_xcap, &natp->nat_ycap);
	repairCoord(&natp->nat_xorg, &natp->nat_yorg);
		
	putnat(natp);
}

void ImportOneNews(void *context, void *object)
{
	Empire_Database *edb = (Empire_Database *)context;
	struct nwsstr *newsp = (struct nwsstr *)object;

	ef_write(EF_NEWS, newsp->nws_uid, newsp);
}

void ImportOneTreaty(void *context, void *object)
{
/*
	just a stub...
*/
}

void ImportOneTrade(void *context, void *object)
{
	Empire_Database *edb = (Empire_Database *)context;
	struct trdstr *tp = (struct trdstr *)object;

	tp->ef_type = EF_TRADE;
	repairCoord(&tp->trd_x, &tp->trd_y);
	puttrade(tp->trd_uid, tp);
}

void ImportOneCommodity(void *context, void *object)
{
	Empire_Database *edb = (Empire_Database *)context;
	struct comstr *cp = (struct comstr *)object;

	cp->ef_type = EF_COMM;
	repairCoord(&cp->com_x, &cp->com_y);
	repairCoord(&cp->sell_x, &cp->sell_y);
	putcomm(cp->com_uid, cp);
}

void ImportOneLoan(void *context, void *object)
{
	Empire_Database *edb = (Empire_Database *)context;
	struct lonstr *lp = (struct lonstr *)object;

	lp->ef_type = EF_LOAN;
	putloan(lp->l_uid, lp);
	printf("importing loan %d\n", lp->l_uid);
}

/*
 * Constants.
 */

char *defaultInput = "csync.dat";

struct DumpTable {
	char *heading;
	EmpireObjectType obj;
	Edb_CallBack dumpfunc;
};

struct DumpTable things[] = {
	{ "Sectors",      V_Sector_Data,   ImportOneSector },
	{ "Nations",      V_Country_Data,  ImportOneNation },
	{ "Ships",        V_Ship_Data,     ImportOneShip },
	{ "Units",        V_Land_Data,     ImportOneUnit },
	{ "Planes",       V_Plane_Data,    ImportOnePlane },
	{ "Nukes",        V_Nuke_Data,     ImportOneNuke },
	{ "News",         V_News_Data,     ImportOneNews },
	{ "Trade",        V_Trade_Data,    ImportOneTrade },
	{ "Commodity",    V_Commodity_Data,ImportOneCommodity },
	{ "Loan",         V_Loan_Data,     ImportOneLoan },
	{ "Treaty",       V_Treaty_Data,   ImportOneTreaty },
	{ 0, 0, NULL} };

/*
 * The import function
 */

process_csync(char *fname)
{
	Empire_Database *edb = 0;
	FILE *f;
	int lines = 0;
	int i;
#define BUFSZ 4096
	char buf[BUFSZ];
	struct DumpTable *p;

	if (!fname) fname = defaultInput;	

	if (!(f= fopen(fname,"r+"))) {
		perror("cannot open input");
		exit(1);
	}
	if (verbose)
		printf("\nReading csync data from %s\n", fname);
	edb = edb_new();
	while (fgets(buf, BUFSZ, f)) {
		edb_parse_c_sync(edb, buf);
		++lines;
		if (verbose) {
			if (lines%10==0)
				putchar('.');
			if (lines%500==0)
				putchar('\n');
		}
	}
	fclose(f);
	if (verbose)
		printf("\nFinished reading csync data...\n");

	for (p=things;p->heading;p++) {
		if (verbose)
			printf("\nImporting %s\n", p->heading);
		edb_enumerate(edb, p->obj, p->dumpfunc, edb);
	}
	if (verbose)
		printf("\nAll Done!\n");
	return 0;
}
